// Author: Michael Mazack - mazack@yahoo.com
// Date: 20100131

// Based on math found at:
// http://mathworld.wolfram.com/WaveEquationRectangle.html

// Standard C++ headers.
#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <ctime>
using namespace std;

// For usleep(). If your system does not have unistd.h or usleep()
// you can work around by simply removing this header and the call
// to usleep(). usleep() will sleep for a specified number of microseconds.
// usleep() is used to regulate the frame rate.
#include <unistd.h>

// OpenGL and GLUT headers.
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

// Constants.
const double pi = acos(-1.0);

//////////////////////////////////////////////

// Mesh related variables.
double **u_mesh;
double **v_mesh;
double *xval;
double *yval;
int xsize, ysize, msize, csize;
double xsmall, ysmall;
double xbig, ybig;
double hx, hy;
double **apq;
double **bpq;
double **opq;


double Lx, Ly;
int p, q;
double h;
double c;

// Tridiagonal system.
double *sup_d;
double *sub_d;
double *diag;
double *rhs;

// Current time and time step.
double t = 0;
double s = 0;
double dt;

// Sleep time between frames in milliseconds.
int ms;

// Dirichlet boundary condition functions.
double xsmall_bc(double);
double ysmall_bc(double);
double xbig_bc(double);
double ybig_bc(double);

// Initial condition.
double initial_cond(double, double);
double neumann_cond(double, double);

//////////////////////////////////////////////

bool wire = false;

// Drawing functions GLUT will need.
void display();
void reshape(int, int);

// Input functions GLUT will need.
void keyboard(unsigned char, int, int);
void mouse_wheel(int, int, int, int);
void mouse_drag(int, int);
void mouse_move(int, int);

void SaveTGA(int);
int shot_num = 0;
int screenshot_mode = 0;

// Variables for old mouse coordinates.
int old_mx;
int old_my;

// Angles and zoom for the molecule.
double xangle = 35.0;
double yangle = 35.0;
double zoom = -30;

int main(int argc, char *argv[])
{
  int i, j;
  double t;

  if(argc < 5)
    {
      cerr << "Usage: wave_equation <x size> <y size> <time step> <delay>" 
	   << endl;
      exit(1);
    }

  // Get the resolution of the mesh and timestep from the command line.
  xsize = atoi(argv[1]) + 2;
  ysize = atoi(argv[2]) + 2;
  dt = (double)atof(argv[3]);
  ms = 1000*atoi(argv[4]);

  csize = 10;

  // Dynamically allocate the mesh.
  u_mesh = new   double* [xsize];
  xval   = new   double  [xsize];
  yval   = new   double  [ysize];
  apq    = new   double* [csize];
  bpq    = new   double* [csize];
  opq    = new   double* [csize];

  for(i = 0; i < xsize; i++)
    u_mesh[i] = new double [ysize];

  for(i = 0; i < csize; i++)
    {
      apq[i]  = new double [csize];
      bpq[i]  = new double [csize];
      opq[i]  = new double [csize];
    }

  // Get the maximum size.
  if(xsize > ysize)
    msize = xsize;
  else
    msize = ysize;

  // Dynamically allocate the tridiagonal system.
  sup_d = new double [msize];
  sub_d = new double [msize];
  diag  = new double [msize];
  rhs   = new double [msize];

  // Initialize parameters.
  xsmall = 0.0;
  ysmall = 0.0;
  xbig   = 10.0;
  ybig   = 10.0;
  hx     = (xbig - xsmall)/(xsize - 1.0);
  hy     = (ybig - ysmall)/(ysize - 1.0);
  t      = 0;
  h      = 0.1;

  Lx = xbig;
  Ly = ybig;

  // Fill in the x vals.
  for(i = 0; i < xsize; i++)
    xval[i] = xsmall + i*hx;

  // Fill in the y vals.
  for(j = 0; j < ysize; j++)
    yval[j] = ysmall + j*hy;

  // x boundary conditions.
  for(i = 0; i < xsize; i++)
    {
      u_mesh[i][0]         = xsmall_bc(yval[i]);
      u_mesh[i][ysize - 1] = xbig_bc(yval[i]);
    }

  // y ends boundary conditions.
  for(j = 0; j < ysize; j++)
    {
      u_mesh[0][j]         = ysmall_bc(xval[j]);
      u_mesh[xsize - 1][j] = ybig_bc(xval[j]);
    }

  // Initial condition.
  for(i = 1; i < xsize - 1; i++)
    for(j = 1; j < ysize - 1; j++)
      u_mesh[i][j] = initial_cond(xval[i], yval[j]); 


  cout << "Computing coefficients ..." << endl;
  // Fill in coefficients.
  for(i = 0; i < csize; i++)
    {
      p = i;
      for(j = 0; j < csize; j++)
	{
	  q = j;

	  opq[p][q] = pi*2*sqrt(((p+1)*(p+1)/(Lx*Lx)) 
				 + ((q+1)*(q+1)/(Ly*Ly)));

	  apq[p][q] = 0;
	  for(s = 0; s < Ly; s += h)
	    {
	      // Evaluate the inner integral for a_{pq}.
	      t = 0; c = 0;
	      for(t = 0; t < Lx; t += h)
		c += initial_cond(t, s)*sin((p+1)*pi*t/Lx);
	      
	      // Multiply by the step size.
	      c *= h;
	      
	      // Now we can calculate a piece of the other
	      // integral.
	      apq[p][q] += c*sin((q+1)*pi*s/Ly);
	    }
	  apq[p][q] *= 4/(Lx*Ly)*h;

	  bpq[p][q] = 0;
	  for(s = 0; s < Ly; s += h)
	    {
	      // Evaluate the inner integral for a_{pq}.
	      t = 0; c = 0;
	      for(t = 0; t < Lx; t += h)
		c += neumann_cond(t, s)*sin((p+1)*pi*t/Lx);
	      
	      // Multiply by the step size.
	      c *= h;
	      
	      // Now we can calculate a piece of the other
	      // integral.
	      bpq[p][q] += c*sin((q+1)*pi*s/Ly);
	    }
	  bpq[p][q] *= 4/(Lx*Ly*opq[p][q])*h;
	
	}
    }
  cout << "Done." << endl;

  // Begin GLUT code.
  glutInit(&argc, argv);

  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
  glutInitWindowSize(512, 512);
  glutCreateWindow("Wave Equation Simulation");

  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyboard);
  glutMouseFunc(mouse_wheel);
  glutMotionFunc(mouse_drag);
  glutPassiveMotionFunc(mouse_move);

  glutDisplayFunc(display);
  glutIdleFunc(display);
  glutMainLoop();
  // End GLUT code.

  // Garbage collection.
  for(i = 0; i < xsize; i++)
    delete [] u_mesh[i];

  for(i = 0; i < csize; i++)
    {
      delete [] apq[i];
      delete [] bpq[i];
      delete [] opq[i];
    }

  delete [] *u_mesh;
  delete [] xval;
  delete [] yval;
  delete [] apq;
  delete [] bpq;

  return 0;
}

double xsmall_bc(double x)
{
  return 0;
}
double ysmall_bc(double y)
{
  return 0;
}

double xbig_bc(double x)
{
  return 0;
}

double ybig_bc(double y)
{
  return 0;
}

double initial_cond(double x, double y)
{
  if(x > 0 && x < 5 && y > 2.5 && y < 8.5)
    return 2.0*sin(x)*sin(y);
  else 
    return 0;
}

double neumann_cond(double x, double y)
{
  return sin(x)*sin(y);
}

void draw_axes()
{
  glLineWidth(5.0);
  glBegin(GL_LINES);
  glColor3f(1.0, 0.0, 0.0);
  glVertex3f(4.0, 0.0, 0.0);
  glVertex3f(0.0, 0.0, 0.0);

  glColor3f(0.0, 1.0, 0.0);
  glVertex3f(0.0, 4.0, 0.0);
  glVertex3f(0.0, 0.0, 0.0);

  glColor3f(0.0, 0.0, 1.0);
  glVertex3f(0.0, 0.0, 4.0);
  glVertex3f(0.0, 0.0, 0.0);
  glEnd();
}

void display()
{
  int i, j, k, l;

  float light_pos[] = { -zoom, -zoom, -zoom, 1.0 };
  float ambient[] = { 0.0, 0.0, 0.0, 1.0 };
  float diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
  float specular[] = { 1.0, 1.0, 1.0, 1.0 };

  double z;

  t += dt;

  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glClearDepth(1);

  glCullFace(GL_BACK);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glEnable(GL_LIGHT0);
  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
  
  glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
  glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, specular);

  glTranslatef(0.0, 0.0, zoom);
  glRotatef(xangle, 0, 1, 0);
  glRotatef(yangle, cos((pi/180)*xangle), 0, sin((pi/180)*xangle)); 

  glLineWidth(1.0);

  glDisable(GL_LIGHTING);

  for(i = 0; i < xsize; i++)
    {
      for(j = 0; j < ysize; j++)
	{
	  z = 0;
	  for(k = 0; k < csize; k++)
	    {
	      for(l = 0; l < csize; l++)
		{
		  z += (apq[k][l]*cos(opq[k][l]*t) + bpq[k][l]*sin(opq[k][l]*t))
		    *sin((k+1)*pi*xval[i]/Lx)*sin((l+1)*pi*yval[j]/Ly);
		}

	    }

	  u_mesh[i][j] = z;
	}
    }

  glBegin(GL_TRIANGLES);
  for(i = 0; i < xsize - 1; i++)
    for(j = 0; j < ysize - 1; j++)
      {
	glColor3f( 0.2, 0.2, fabs(0.8*((2.0 + u_mesh[i][j])/4.0)) + 0.2);

	// Draw the first face. X Z Y
	glVertex3d(xval[i] - (xbig - xsmall)/2,
		   u_mesh[i][j], 
		   (ybig - yval[j]) - (ybig - ysmall)/2);

	glVertex3d(xval[i + 1] - (xbig - xsmall)/2,
		   u_mesh[i + 1][j], 
		   (ybig - yval[j]) - (ybig - ysmall)/2);

	glVertex3d(xval[i + 1] - (xbig - xsmall)/2,
		   u_mesh[i + 1][j + 1], 
		   (ybig - yval[j + 1]) - (ybig - ysmall)/2);

	// Draw the second face. X Z Y
	glVertex3d(xval[i] - (xbig - xsmall)/2,
		   u_mesh[i][j], 
		   (ybig - yval[j]) - (ybig - ysmall)/2);

	glVertex3d(xval[i + 1] - (xbig - xsmall)/2,
		   u_mesh[i + 1][j + 1], 
		   (ybig - yval[j + 1]) - (ybig - ysmall)/2);

	glVertex3d(xval[i] - (xbig - xsmall)/2,
		   u_mesh[i][j + 1], 
		   (ybig - yval[j + 1]) - (ybig - ysmall)/2);
      }

  glEnd();


  glTranslatef(-20, -20, -30);

  glRotatef(xangle, 0, 1, 0);
  glRotatef(yangle, cos((pi/180)*xangle), 0, sin((pi/180)*xangle));

  glLoadIdentity();

  //  draw_axes();

  glutSwapBuffers();

  if(screenshot_mode)
    {
      SaveTGA(shot_num);
      shot_num++;
    }

  usleep(ms);
}

void reshape(int w, int h)
{
  glViewport(0, 0, w, h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();


  glLineWidth(5.0);

  glEnable(GL_DEPTH_TEST);
  glEnable(GL_BLEND);
  glEnable(GL_LIGHTING);

  glShadeModel(GL_SMOOTH);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  gluPerspective(30.0f, (double)w/(double)h, 2.0, 5000.0);
  //glOrtho(-30, 30, -30, 30, 20, 2000.0);
  
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}

void keyboard(unsigned char key, int x, int y)
{
  switch(key)
    {
    case 27: /* Escape Key */
      exit(0);
      break;

    case '+':
    case '=':
      zoom += 5;
      break;

    case '-':
    case '_':
      zoom -= 5;
      break;

    case 'a':
      xangle += 0.5;
      break;

    case 'd':
      xangle -= 0.5;
      break;

    case 's':
      yangle -= 0.5;
      break;

    case 'w':
      yangle += 0.5;
      break;

    case 'f':
      if(wire == true)
	{
	  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	  wire = false;
	}
      else
	{
	  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	  wire = true;
	}
      break;
      

    case 'c':
      if(screenshot_mode)
	screenshot_mode = 0;
      else
	screenshot_mode = 1;
      break;

    default:
      break;
    }
}

void mouse_wheel(int button, int state, int x, int y)
{
  if(state == GLUT_UP)
    {
      if(button == 3) // Mouse wheel up.
	{
	  zoom -= 5;
	}
      else if(button == 4) // Mouse wheel down.
	{
	  zoom += 5;
	}
    }
}

void mouse_drag(int x, int y)
{
  xangle += -(old_mx - x)*0.02;
  yangle += -(old_my - y)*0.02; 

}

void mouse_move(int x, int y)
{
  old_mx = x;
  old_my = y;
}

void SaveTGA(int nShot)
{

    glPixelStorei(GL_PACK_ALIGNMENT,1);
      
    char cFileName[64];
    FILE *fScreenshot;
    int nSize = 512 * 512 * 3;
    
    GLubyte *pixels = new GLubyte [nSize];
    if (pixels == NULL) 
      return;    

    sprintf(cFileName,"frame_%06d.tga",nShot);
        
    fScreenshot = fopen(cFileName,"wb");
    
    glReadPixels(0, 0, 512, 512, GL_RGB, GL_UNSIGNED_BYTE, pixels);
           
    //convert to BGR format    
    unsigned char temp;
    int i = 0;
    while (i < nSize)
    {
        temp = pixels[i];       //grab blue
        pixels[i] = pixels[i+2];//assign red to blue
        pixels[i+2] = temp;     //assign blue to red

        i += 3;     //skip to next blue byte
    }

    unsigned char TGAheader[12]={0,0,2,0,0,0,0,0,0,0,0,0};
    unsigned char header[6] = {512%256, 512/256, 512%256, 512/256, 24, 0};
    
    fwrite(TGAheader, sizeof(unsigned char), 12, fScreenshot);
    fwrite(header, sizeof(unsigned char), 6, fScreenshot);
    fwrite(pixels, sizeof(GLubyte), nSize, fScreenshot);
    fclose(fScreenshot);
        
    delete [] pixels;   
    
    return;
}
